/* grub-verify.c - debug tool for verify GRUB module */
/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 2008,2009,2010,2015 Free Software Foundation, Inc.
 *
 *  GRUB is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>
#include <grub/types.h>
#include <grub/util/misc.h>
#include <grub/emu/hostdisk.h>
#include <grub/crypto.h>
#include <grub/file.h>
#include <grub/pubkey.h>
#include <grub/env.h>
#include <grub/command.h>

#include <stdio.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include "progname.h"
#pragma GCC diagnostic ignored "-Wmissing-prototypes"
#pragma GCC diagnostic ignored "-Wmissing-declarations"
#include "argp.h"
#pragma GCC diagnostic error "-Wmissing-prototypes"
#pragma GCC diagnostic error "-Wmissing-declarations"

static const char *Myname;

static void
usage(void)
{
	fprintf(stderr, "Usage: %s [-d] [-l] [-k pubkey] [-o object] [-s sig]\n",
	    Myname);
	exit(1);

}

static grub_err_t
verify (const char *kflag, const char *oflag, const char *sflag)
{
	grub_file_t pkf, f, sig;
	struct grub_public_key *pk = NULL;
	grub_err_t err = GRUB_ERR_NONE;

	f = sig = pkf = NULL;

	grub_dprintf ("verify", "attempting to open pkf\n");

	grub_file_filter_disable_all ();
	pkf = grub_file_open (kflag);
	if (!pkf) {
		grub_util_error ("unable to open pubkey: %s\n", kflag);
		err = grub_errno;
		goto out;
	}

	grub_dprintf ("verify", "opened pkf\n");

	pk = grub_load_public_key (pkf);
	if (!pk)
	{
		grub_util_error ("unable to parse pubkey: %s\n", kflag);
		err = grub_errno;
		goto out;
	}

	grub_dprintf ("verify", "loaded pk\n");

	grub_file_filter_disable_all ();
	f = grub_file_open (oflag);
	if (!f) {
		grub_util_error ("unable to open object file: %s\n", oflag);
		err = grub_errno;
		goto out;
	}

	grub_dprintf ("verify", "openning object\n");

	grub_file_filter_disable_all ();
	sig = grub_file_open (sflag);
	if (!sig) {
		grub_util_error ("unable to open signature file: %s\n", sflag);
		err = grub_errno;
		goto out;
	}

	grub_dprintf ("verify", "openning signature\n");

	err = grub_verify_signature (f, sig, pk);

	grub_dprintf ("verify", "verifying siganture: %d\n", err);

out:
	if (f)
		grub_file_close(f);

	if (sig)
		grub_file_close(sig);

	if (pkf)
		grub_file_close (pkf);


	return err;

}

static grub_err_t
list_fingerprint(char *kflag)
{
	grub_command_t cmd;
	char *args[] = { kflag };
	const char *name = "trust";
	grub_err_t err = GRUB_ERR_NONE;

	cmd = grub_command_find (name);
	if (! cmd)
		grub_util_error (_("can't find command `%s'"), name);

	err = (cmd->func) (cmd, 1, args);
	if (err) {
		grub_util_error (_("can't invoke command `%s': %d"), name, err);
	}

	name = "list_trusted";
	cmd = grub_command_find (name);
	if (! cmd)
		grub_util_error (_("can't find command `%s'"), name);

	err = (cmd->func) (cmd, 0, NULL);
	if (err) {
		grub_util_error (_("can't invoke command `%s'"), name);
	}

	return err;
}

static char *
grub_canonicalize(const char *p)
{
	char *canon, *r;

	canon = NULL;

	if (p[0] == '/') {
		asprintf(&canon, "(host)%s", p);

	} else if (p[0] == '(') {

		return strdup(p);
	} else {

		r = realpath(p, NULL);
		if (r) {
			asprintf(&canon, "(host)%s", r);
			free(r);
		} else {
			return strdup(p);
		}
	}

	return canon;
}

int
main (int argc, char *argv[])
{
	int opt, lflag, dflag;
	char *kflag, *oflag, *sflag;
	grub_err_t err = GRUB_ERR_NONE;

	dflag = lflag = 0;
	sflag = oflag = kflag = NULL;

	Myname = argv[0];

	grub_util_host_init (&argc, &argv);

	while ((opt = getopt(argc, argv, "dlk:o:s:")) != -1) {
		switch (opt) {
		case 'd':
			dflag = 1;
			break;
		case 'l':
			lflag = 1;
			break;
		case 'k':
			if ((kflag = grub_canonicalize(optarg)) == NULL) {
				usage();
			}
			break;
		case 'o':
			if ((oflag = grub_canonicalize(optarg)) == NULL) {
				usage();
			}
			break;
		case 's':
			if ((sflag = grub_canonicalize(optarg)) == NULL) {
				usage();
			}
			break;
		default: /* '?' */
			usage();
			break;
		}
	}

	if (dflag)
		grub_env_set ("debug", "disk verify crypt");

	/* Initialize all modules. */
	grub_init_all ();
	grub_gcry_init_all ();
	grub_hostfs_init();
	grub_host_init();
	grub_verify_init();

	if (lflag) {
		if (kflag == NULL) {
			usage();
		}
		list_fingerprint((char *)kflag);
	} else {
		if (kflag == NULL || oflag == NULL || sflag == NULL) {
			usage();
		}
		err = verify(kflag, oflag, sflag);
		grub_dprintf ("verify", "verify ret: %d\n", err);
	}


	/* Free resources.  */
	grub_gcry_fini_all ();
	grub_hostfs_fini();
	grub_host_fini();
	grub_fini_all ();
	grub_verify_fini ();

	if (sflag)
		free(sflag);

	if (oflag)
		free(oflag);

	if (kflag)
		free(kflag);
	return err;
}
